/* $Id: natpmp.c,v 1.3 2007/10/25 13:08:44 nanard Exp $ */
/* MiniUPnP project
 * (c) 2007 Thomas Bernard
 * http://miniupnp.free.fr/ or http://miniupnp.tuxfamily.org/
 * This software is subject to the conditions detailed
 * in the LICENCE file provided within the distribution */
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "config.h"
#include "natpmp.h"
#include "upnpglobalvars.h"
#include "getifaddr.h"
#include "upnpredirect.h"

#ifdef ENABLE_NATPMP

/* TODO : clean that ! */
extern int
get_redirect_rule(const char * ifname, unsigned short eport, int proto,
                  char * iaddr, int iaddrlen, unsigned short * iport,
                  char * desc, int desclen,
                  u_int64_t * packets, u_int64_t * bytes);

void ProcessIncomingNATPMPPacket(int s)
{
	unsigned char req[32];	/* request udp packet */
	unsigned char resp[32];	/* response udp packet */
	int resplen;
	struct sockaddr_in senderaddr;
	socklen_t senderaddrlen = sizeof(senderaddr);
	int n;
	char tmp[16];
	char senderaddrstr[16];
	n = recvfrom(s, req, sizeof(req), 0,
	             (struct sockaddr *)&senderaddr, &senderaddrlen);
	if(n<0) {
		syslog(LOG_ERR, "recvfrom(natpmp): %m");
		return;
	}
	inet_ntop(AF_INET, &senderaddr.sin_addr,
	          senderaddrstr, sizeof(senderaddrstr));
	syslog(LOG_INFO, "NAT-PMP request received from %s:%hu",
           senderaddrstr, ntohs(senderaddr.sin_port));
           //inet_ntoa(senderaddr.sin_addr), ntohs(senderaddr.sin_port));
	if(n<2) {
		syslog(LOG_WARNING, "discarding NAT-PMP request (too short)");
		return;
	}
	memset(resp, 0, sizeof(resp));
	resplen = 8;
	resp[1] = 128 + req[1];	/* response OPCODE is request OPCODE + 128 */
	*((uint32_t *)(resp+4)) = htonl(time(NULL) - startup_time);
	if(req[0] > 0) {
		/* invalid version */
		syslog(LOG_WARNING, "unsupported NAT-PMP version : %u",
		       (unsigned)req[0]);
		resp[3] = 1;	/* unsupported version */
	} else {
		switch(req[1]) {
		case 0:	/* Public address request */
			syslog(LOG_INFO, "NAT-PMP public address request");
			if(use_ext_ip_addr) {
                inet_pton(AF_INET, use_ext_ip_addr, resp+8);
			} else {
				if(getifaddr(ext_if_name, tmp, INET_ADDRSTRLEN) < 0) {
					syslog(LOG_ERR, "Failed to get IP for interface %s", ext_if_name);
				}
				inet_pton(AF_INET, tmp, resp+8);
			}
			resplen = 12;
			break;
		case 1:
		case 2:
			{
				unsigned short iport;	/* private port */
				unsigned short eport;	/* public port */
				uint32_t lifetime;
				int r;
				int proto;
				char iaddr_old[16];
				unsigned short iport_old;
				iport = ntohs(*((uint16_t *)(req+4)));
				eport = ntohs(*((uint16_t *)(req+6)));
				lifetime = ntohl(*((uint32_t *)(req+8)));
				proto = (req[1]==1)?IPPROTO_TCP:IPPROTO_UDP;
				syslog(LOG_INFO, "NAT-PMP port mapping request : "
				                 "%hu->%s:%hu %s lifetime=%us",
				                 eport, senderaddrstr, iport,
				                 (req[1]==1)?"tcp":"udp", lifetime);
				if(eport==0)
					eport = iport;
				if(iport==0
				   || !check_upnp_rule_against_permissions(upnppermlist, num_upnpperm, eport, senderaddr.sin_addr, iport)) {
					resp[3] = 2;	/* Not Authorized/Refused */
				} else do {
					r = get_redirect_rule(ext_if_name, eport, proto,
					                      iaddr_old, sizeof(iaddr_old),
					                      &iport_old, 0, 0, 0, 0);
					if(r==0) {
						if(strcmp(senderaddrstr, iaddr_old)==0
					       && iport==iport_old) {
							/* redirection allready ok */
							syslog(LOG_INFO, "port %hu %s already redirected to %s:%hu", eport, (proto==IPPROTO_TCP)?"tcp":"udp", iaddr_old, iport_old);
							break;
						} else {
							eport++;
						}
					} else {
						/* do the redirection */
						if(lifetime>0) {
							char desc[64];
							snprintf(desc, sizeof(desc), "NAT-PMP %u",
							         (unsigned)(time(NULL) - startup_time)
							         + lifetime);
							/* TODO : check return code */
							upnp_redirect_internal(eport, senderaddrstr, iport,
							                       proto, desc);
						}
					}
				} while(r==0);
				*((uint16_t *)(resp+8)) = iport;	/* private port */
				*((uint16_t *)(resp+10)) = eport;	/* public port */
				*((uint32_t *)(resp+12)) = lifetime;
			}
			resplen = 16;
			break;
		default:
			resp[3] = 5;	/* Unsupported OPCODE */
		}
	}
	n = sendto(s, resp, resplen, 0,
	           (struct sockaddr *)&senderaddr, sizeof(senderaddr));
}

#endif

